# -*- coding:UTF-8 -*-
import math
import os
import sys
import psutil
import warnings

import pandas as pd
import pymongo
import requests


def in_severs() -> bool :
    loc: str=os.getenv('SERVER_LOCATION')
    if (loc=='CN' or loc is None) and sys.platform.startswith('linux') :
        return True
    else :
        return False


def get_mongodb() :
    return os.environ.get('MONGODB_URL')


def get_db_info(db: object) -> object :
    db_evn=get_mongodb()
    bank_client=pymongo.MongoClient(db_evn)
    bank_col=bank_client['db-info']['raw_db']
    db_info=bank_col.find_one({'_id' : db})
    if not db_info :
        db_info=bank_col.find_one({'_id' : '_key_'})
        db_info['db']=db
        del db_info['_id']
    return db_info


def df_creator(data, index=None, delitem=None) :
    if delitem is None :
        delitem=[]
    if index :
        df=pd.DataFrame(data, index=index)
    else :
        df=pd.DataFrame(data)
    for i in delitem :
        del df[i]
    return df


def _dblink() :
    db_evn=get_mongodb()
    bank_client=pymongo.MongoClient(db_evn)
    bank_col=bank_client['db-info']['raw_db']
    return bank_col


def dblink(db, col) :
    db_info=_dblink().find_one({'_id' : db})
    client=pymongo.MongoClient(db_info['uri'])
    db=client[db_info['db']]
    collection=db[col]
    return collection


def col_stats(db, col) :
    db_info=_dblink().find_one({'_id' : db})
    client=pymongo.MongoClient(db_info['uri'])
    db=client[db_info['db']]
    return db.command('collstats', col)


def db_stats(db) :
    db_info=_dblink().find_one({'_id' : db})
    client=pymongo.MongoClient(db_info['uri'])
    db=client[db_info['db']]
    return db.command('dbstats')


def dblink_help(db=False) :
    if db :
        db_info=_dblink().find_one({'_id' : db})
        client=pymongo.MongoClient(db_info['uri'])
        db=client[db_info['db']]
        return db.collection_names()
    else :
        dbs=_dblink().find()
        return [i['_id'] for i in dbs]


def df2mongo(df, col) :
    data=df.to_dict('records')
    col.insert_many(data)


def mongo2df(col, match=None, projection=None, sort=None, skip=0, limit=0) :
    return df_creator(list(col.find(filter=match, projection=projection, sort=sort, skip=skip, limit=limit)))


def collection_show(db, col) :
    return mongo2df(dblink(db, col), limit=100)


def dblink_add(_id, uri, db=False) :
    _dblink().insert({'_id' : _id, 'uri' : uri, 'db' : db if db is not False else uri.split('/')[-1]})
    return _id in dblink_help()


def dblink_remove(_id) :
    _dblink().remove({'_id' : _id})


def dblink_update(_id, uri, db=False) :
    _dblink().update({'_id' : _id}, {'uri' : uri, 'db' : db if db is not False else uri.split('/')[-1]})


def get_token(job, on) :
    db_evn=get_mongodb()
    bank_client=pymongo.MongoClient(db_evn)
    bank_col=bank_client['tk'][job]
    tk=bank_col.find_one({'on' : on})
    return tk['token'], tk['servers']


def crawler_starter(db: str, col: str, work_on: str) :
    db=get_db_info(db)['db']
    token, servers=get_token('crawler_starter', work_on.upper())
    body={"db" : db, "table" : col, "token" : token}
    return requests.post(url=servers, json=body)



import pickle
from sys import getsizeof


def easy_dump(obj, file_name):
    """
    将对象以pickle格式保存到文件中。

    参数:
        obj: 要保存的对象。
        file_name: 保存的文件名，如果没有以'.pkl'结尾，会自动添加后缀。

    返回:
        无。
    """
    if not file_name.endswith('.pkl'):
        file_name += '.pkl'
    with open(file_name, 'wb') as file:
        pickle.dump(obj, file)


def easy_load(file_name: str):
    """
    从pickle文件中加载对象。

    参数:
        file_name: 要加载的文件名，如果没有以'.pkl'结尾，会自动添加后缀。

    返回:
        加载的对象。
    """
    if not file_name.endswith('.pkl'):
        file_name += '.pkl'
    with open(file_name, 'rb') as file:
        obj = pickle.load(file)
        print('object type:', type(obj))
        print('object size: ', getsizeof(obj))
        return obj



# -*- coding:UTF-8 -*-
# ANBUtils/DBWorker.py
# Created by redbson on 2021-09-01
# Latest modified on 2023-03-28

# This module provides a MongoDB database worker class for data operations. The class provides methods for connecting
# to a MongoDB database, retrieving data from collections, and inserting data into collections. The class also
# provides methods for auditing collections and the database. The class can be used to perform various data
# operations on MongoDB databases.


import hashlib
import math
import os
import psutil
import time
import warnings
from typing import List, Dict, Callable
from .db_tools import in_severs, get_mongodb, get_db_info, df_creator
import pymongo
import pandas as pd


class DBWorker(object):
    def __init__(self, db):
        """
        Initialize a MongoDB database worker.

        Args:
            db (str): The name of the database.
        """
        db_info: dict = get_db_info(db)
        public_uri: str = os.environ.get('MONGODB_PUB_URI')
        uri: str = db_info['uri'] if in_severs() else (db_info['uri'].split(
            '@')[0] + '@' + public_uri + ':' + db_info['uri'].split(':')[-1])
        self.db_code: str = db
        self.client: pymongo.MongoClient = pymongo.MongoClient(uri)
        self.db: pymongo.database.Database = self.client[db_info['db']]
        self.col: List[Dict[str, pymongo.collection.Collection]] = {}
        self._link('__audit__')
        self._link('__log__')

    def _link(self, col: str) -> None:
        if col not in self.col:
            self.col[col] = self.db[col]

    def link(self, col: str) -> pymongo.collection.Collection:
        """
        Get a reference to the specified collection in the database.

        Args:
            col (str): The name of the collection.

        Returns:
            pymongo.collection.Collection: The collection object.
        """
        self._link(col)
        return self.col[col]

    def _last_audit(self, col_name) -> dict or None:
        return self.link('__audit__').find_one(
            {"collection": col_name}, sort = [("audit_ts", -1)])

    def _audit(self, col_name: str) -> bool:
        def quick_audit_col(_col_name: str, _factor: str = '') -> dict:
            _version: str = '0.0.1'
            time_stamp: int = time.time_ns()
            stats: dict = self.get_col_stats(col_name)
            stream: str = (
                f"{stats['ns']}:{stats['size']}#{stats['count']}#{stats['size']}#{stats['storageSize']}#"
                f"{stats['nindexes']}#{stats['totalIndexSize']}")
            hash_result: str = hashlib.sha256(
                stream.encode('utf-8')).hexdigest()
            _id: str = hashlib.sha256(
                f"{hash_result}:{time_stamp}@{factor}".encode('utf-8')).hexdigest()
            return {
                "_id": _id,
                "collection": col_name,
                "ns": stats['ns'],
                "hash": hash_result,
                "document_count": stats['count'],
                "data_size": stats['size'],
                "storage_size": stats['storageSize'],
                "index_count": stats['nindexes'],
                "total_index_size": stats['totalIndexSize'],
                "average_object_size": stats['avgObjSize'] if 'avgObjSize' in stats else 'N/A',
                "raw_data": stats,
                "audit_time_stamp": time_stamp,
                "meta": {
                    "method": "quick_audit_collection",
                    "version": _version}}

        last_report: dict or None = self._last_audit(col_name)
        factor: str = last_report['hash'] if last_report is not None else ''
        new_report: dict = quick_audit_col(col_name, factor)

        if last_report is None:
            self.link('__audit__').insert_one(new_report)
        elif last_report['hash'] != new_report['hash']:
            self.link('__audit__').insert_one(new_report)
            return False
        return True

    def get_col_stats(self, col: str) -> dict:
        """
            Get the statistics of a collection in the database.

            Args:
                col (str): The name of the collection.

            Returns:
                dict: The statistics of the collection.
        """
        self._link(col)
        return self.db.command('collstats', col)

    def get_db_stats(self) -> dict:
        """
              Get the statistics of the database.

              Returns:
                  dict: The statistics of the database.
        """
        return self.db.command('dbstats')

    def get_cols_name(self) -> List[str]:
        """
              Get the names of all collections in the database.

              Returns:
                  list: The names of the collections.
        """
        return self.db.list_collection_names()

    def _to_df_base(
            self,
            col: str,
            match: dict = None,
            projection: dict = None,
            sort: dict = None,
            skip: dict = 0,
            limit: dict = 0) -> pd.DataFrame:
        self._link(col)
        return df_creator(
            list(
                self.col[col].find(
                    filter = match,
                    projection = projection,
                    sort = sort,
                    skip = skip,
                    limit = limit)))

    def _to_df_large(
            self,
            col: str,
            match: dict = None,
            projection: dict = None,
            verbose = True) -> pd.DataFrame:
        """
            Convert a large collection into a DataFrame by splitting it into multiple parts.

            Args:
                col (str): The name of the collection.
                match (dict, optional): The query filter. Defaults to None.
                projection (dict, optional): The projection query. Defaults to None.
                verbose (bool, optional): Whether to display progress messages. Defaults to True.

            Returns:
                pandas.DataFrame: The DataFrame created from the collection data.
        """

        self._link(col)
        col_status: dict = self.get_col_stats(col)
        st_size: float = col_status['size'] / 1024 ** 2
        st_count: int = col_status['count']
        n: int = math.ceil(st_size / 200)
        step: int = math.ceil(st_count / n)

        if n > 1:
            if verbose:
                print(
                    f'{col} has {st_count:,d} records, {st_size:,.2f} MB, split to {n} parts')
            mem: int = psutil.virtual_memory()
            if mem.total < col_status['size']:
                raise MemoryError(
                    'Not enough memory to process {col}'.format(
                        col = col))

            if mem.available < col_status['size']:
                warnings.warn(
                    'Please note that there may be insufficient memory when reading the {col} in the current system '
                    'environment.'.format(
                        col = col))

            dfs: List[pd.DataFrame] = []

            for i in range(n):
                if verbose:
                    print('processing {i:>2d} part ...'.format(i = i + 1))
                _df: pd.DataFrame = self._to_df_base(
                    col, match = match, projection = projection, skip = i * step, limit = step)
                dfs.append(_df)
            df: pd.DataFrame = pd.concat(dfs)
            if verbose:
                print('done')

        else:
            if verbose:
                print(
                    '{col} has {count:,d} records, {size:,.2f} MB, processing ...'.format(
                        col = col, count = st_count, size = st_size))
            df: pd.DataFrame = self._to_df_base(
                col, match = match, projection = projection)

        return df

    def to_df(
            self,
            col: str,
            match: dict = None,
            projection: dict = None,
            sort: str = None,
            skip: int = 0,
            limit: int = 0) -> pd.DataFrame:
        """
        Convert the data from a collection into a DataFrame.

        Args:
            col (str): The name of the collection.
            match (dict, optional): The query filter. Defaults to None.
            projection (dict, optional): The projection query. Defaults to None.
            sort (list, optional): The sort order. Defaults to None.
            skip (int, optional): The number of documents to skip. Defaults to 0.
            limit (int, optional): The maximum number of documents to return. Defaults to 0.

        Returns:
            pandas.DataFrame: The DataFrame created from the collection data.
        """

        if not any((match, projection, sort, limit, skip)):
            return self._to_df_large(col)
        else:
            return self._to_df_base(col, match, projection, sort, skip, limit)

    def to_df_many(
            self,
            cols: List[str],
            match: dict = None,
            projection: dict = None) -> pd.DataFrame:
        """
         Convert multiple collections into a single DataFrame.

         Args:
             cols (list): The names of the collections.
             match (dict, optional): The query filter. Defaults to None.
             projection (dict, optional): The projection query. Defaults to None.

         Returns:
             pandas.DataFrame: The DataFrame created from the collections' data.
         """
        dfs: List[pd.DataFrame] = [self.to_df(
            c, match = match, projection = projection) for c in cols]
        df: pd.DataFrame = pd.concat(dfs)
        df.reset_index(inplace = True, drop = True)
        return df

    def insert_df(self, df: pd.DataFrame, col: str) -> None:
        """
          Insert a DataFrame into a collection.

          Args:
              df (pandas.DataFrame): The DataFrame to be inserted.
              col (str): The name of the collection.
        """
        self._link(col)
        data: List[dict] = df.to_dict('records')
        self.col[col].insert_many(data)

    def update_df(self, df: pd.DataFrame, col: str, key: str) -> None:
        """
            Update documents in a collection using data from a DataFrame.

            Args:
                df (pandas.DataFrame): The DataFrame containing the updated data.
                col (str): The name of the collection.
                key (str): The key field used to match and update documents.
        """
        # 创建一个空的批量操作列表
        bulk_operations = []

        # 为每条记录创建一个UpdateOne操作，并添加到批量操作列表中
        for record in data:
            filter = {key: record[key]}  # 定位需要更新的文档
            update = {'$set': record}  # 定义更新操作
            bulk_operations.append(pymongo.UpdateOne(filter, update, upsert = True))

        # 执行批量操作
        if bulk_operations:  # 确保有操作要执行
            result = self.col[col].bulk_write(bulk_operations)
            print(f"Bulk operation completed. Matched: {result.matched_count}, "
                  f"Modified: {result.modified_count}, "
                  f"Upserted: {result.upserted_count}")
        else:
            print("No operations to perform.")

    def collection_sample(
            self,
            col: str,
            sample_size: int = 100) -> pd.DataFrame:
        """
        Get a sample of documents from a collection.

        Args:
            col (str): The name of the collection.
            sample_size (int, optional): The number of documents to retrieve. Defaults to 100.

        Returns:
            pandas.DataFrame: The sampled data as a DataFrame.
        """
        return self.to_df(col, limit = sample_size)

    def data_insert(self, col: str, data: dict or list) -> None:
        """
        Insert data into a collection.

        Args:
            col (str): The name of the collection.
            data (dict): The data to be inserted.
        """
        self._link(col)
        _func: Callable[[str], object] = self.col[col].insert_one if isinstance(
            data, dict) else self.col[col].insert_many
        _func(data)

    def drop_col(self, col: str) -> None:
        """
        Drop a collection from the database.

        Args:
            col (str): The name of the collection to be dropped.
        """
        self._link(col)
        self.col[col].dorp()

    def audit_many(self, cols: List[str]) -> List[str]:
        """
        Audit multiple collections in the database.

        Args:
            cols (list): The names of the collections.

        Returns:
            list: The names of the collections that have not pass audited.
        """
        return [col_name for col_name in cols if not self._audit(col_name)]

    def audit_one(self, col: str) -> bool:
        """
        Audit a collection in the database.

        Args:
            col (str): The name of the collection.

        Returns:
            bool: True if the collection has passed the audit, False otherwise.
        """
        return self._audit(col)

    def audit_db(self) -> List[str]:
        """
        Audit all collections in the database.

        Returns:
            list: The names of the collections that have not pass audited.
        """
        col_names: List[str] = self.get_cols_name()
        col_names.remove('__audit__')
        col_names.remove('__log__')
        col_names.remove('system.indexes')
        return self.audit_many(col_names)

    def logging(self, data: dict, _type = '') -> None:
        """
        Insert a log entry into the database.

        Args:
            data (dict): The log data to be inserted.
        """
        data['_type_'] = type
        self.link('__log__').insert_one(data)

    def read_logging(self, _type: str) -> pd.DataFrame:
        return self.link('__log__').find_all({"_type_": _type})

    def __del__(self):
        self.client.close()

    def __repr__(self):
        return f"DBWorker({self.db_code})"



from .environ import environment_checker as _checker

_checker()

from .environ import (
    set_environ
)

from .a import (
    print_rate_progress, set_date_index, digit, count, value, count2int, data_browser
)

from .db_tools import (
    crawler_starter, dblink, dblink_add, dblink_remove, dblink_update, collection_show,
    df2mongo, mongo2df, get_db_info, get_mongodb, dblink_help, get_token
)

from .DBWorker import (
    DBWorker
)

from .easy_pickle import (
    easy_dump, easy_load
)

from .id_work import (
    int_mark, id_analyst, matplot_set
)

from .messenger import (
    qywechat_message, qywechat_text_message, dingtalk_message, dingtalk_text_message, message_mark_A, message_mark_B,
    message_mark_C, message_mark_D
)

from .tbox import (
    future, future_base, date_format, today, yesterday, tomorrow, ts2str, now, utc2tz, previous_date
)



from setuptools import setup, find_packages

with open('README.md', 'r', encoding = 'utf-8') as f:
    long_description = f.read()

setup(
    name = "ANBUtils",
    version = '1.7.3',
    packages = find_packages(),
    author = 'Yafei Hou',
    author_email = 'redbson@gmail.com',
    url = 'https://github.com/redbson/ANBUtils',
    description = 'ANBUilts is a versatile Python package that offers a comprehensive set of utility functions and '
                  'tools for data analysis, database operations, and messaging integration.',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    python_requires = '>=3.8',
    install_requires = [
        "matplotlib>=3.0.0",
        "numpy>=1.0.0",
        "pandas>=1.0.0",
        "pymongo>=4.0.2",
        "requests>=2.0.0",
        "matplotlib>=3.0.0",
        "psutil>=5.0.0",
    ]
)



# -*- coding:UTF-8 -*-

import time
from datetime import datetime, timedelta


def future(x):
    """
    计算未来日期。

    根据给定的天数偏移量，计算未来的日期。

    参数：
        x (int): 天数偏移量。

    返回：
        str: 未来日期的字符串表示（格式：YYYY-MM-DD）。

    异常：
        None
    """
    return (datetime.now() + timedelta( days=x )).strftime( '%Y-%m-%d' )


def utc2tz(dt, tz='E8'):
    """
    将UTC时间转换为指定时区的时间。

    根据给定的UTC时间和时区偏移量，将UTC时间转换为指定时区的时间。

    参数：
        dt (datetime): UTC时间。
        tz (str): 时区偏移量（默认为'E8'，表示东八区）。

    返回：
        datetime: 指定时区的时间。

    异常：
        ValueError: 当时区偏移量不在'E0'至'W12'范围内时，引发该异常。
    """
    if tz[0].upper() == 'E':
        h = int( tz[1:] )
    elif tz[0].upper() == 'W':
        h = int( tz[1:] ) * -1
    else:
        raise ValueError( '[utc2tz] tz is KeyError!' )

    return dt + timedelta( hours=h )


def today():
    """
    获取当前日期。

    返回：
        str: 当前日期的字符串表示（格式：YYYY-MM-DD）。

    异常：
        None
    """
    return datetime.now().strftime( "%Y-%m-%d" )


def tomorrow():
    """
    获取明天的日期。

    返回：
        str: 明天的日期的字符串表示（格式：YYYY-MM-DD）。

    异常：
        None
    """
    return future( 1 )


def yesterday():
    """
    获取昨天的日期。

    返回：
        str: 昨天的日期的字符串表示（格式：YYYY-MM-DD）。

    异常：
        None
    """
    return future( -1 )


def now():
    """
    获取当前日期和时间。

    返回：
        str: 当前日期和时间的字符串表示（格式：YYYY-MM-DD HH:MM）。

    异常：
        None
    """
    return datetime.now().strftime( '%Y-%m-%d %H:%M' )


def future_base(date, x):
    """
    根据基准日期计算未来日期。

    根据给定的基准日期和天数偏移量，计算未来的日期。

    参数：
        date (str): 基准日期的字符串表示（格式：YYYY-MM-DD）。
        x (int): 天数偏移量。

    返回：
        str: 未来日期的字符串表示（格式：YYYY-MM-DD）。

    异常：
        None
    """
    base = datetime.strptime( date, "%Y-%m-%d" )
    return (base + timedelta( days=x )).strftime( '%Y-%m-%d' )


def ts2str(x):
    """
    将时间戳转换为日期字符串。

    根据给定的时间戳，将其转换为日期字符串（格式：YYYY-MM-DD）。

    参数：
        x (int or str): 时间戳。

    返回：
        str: 日期字符串（格式：YYYY-MM-DD）。

    异常：
        None
    """
    if type( x ) == str:
        x = int( x )
    return time.strftime( "%Y-%m-%d", time.localtime( x ) )


def date_format(date, date_format='YYYY-MM-DD'):
    """
    格式化日期字符串。

    根据给定的日期字符串和日期格式，将日期字符串格式化为指定的日期格式。

    参数：
        date (str): 日期字符串。
        date_format (str): 日期格式（默认为'YYYY-MM-DD'）。

    返回：
        str: 格式化后的日期字符串。

    异常：
        None
    """
    if len( date ) == 10:
        if len( date.split( '-' ) ) == 3:
            date = date
        elif len( date.split( '/' ) ) == 3:
            date = date.replace( '/', '-' )
        elif len( date.split( '_' ) ) == 3:
            date = date.replace( '_', '-' )

    if len( date ) == 6:
        y = date[:2]
        m = date[2:4]
        d = date[4:]
        if int( y ) <= 30:
            y = '20' + y
        else:
            y = '19' + y
        date = '-'.join( [y, m, d] )

    if date_format == 'YYYY-MM-DD':
        return date
    elif date_format == 'YYMMDD':
        return date.replace( '-', '' )[2:]
    elif date_format == 'YYYY_MM_DD':
        return date.replace( '-', '_' )


def previous_date(n, base_day=None, mode='m'):
    if base_day is None:
        base_day = datetime.today()
    else:
        base_day = datetime.strptime( base_day, '%Y-%m-%d' )

    if mode == 'm':
        pass

    elif mode == 'q':
        current_month = base_day.month
        base_day = datetime( base_day.year, ((current_month - 1) // 3) * 3 + 1, 1 )

    else:
        raise ValueError( 'mode must be m or q' )

    # 获取当前月份的年份和月份
    current_year, current_month = base_day.year, base_day.month

    # 计算当前月份之前的n个月
    n = n if mode == 'm' else n * 3
    for _ in range( n ):
        # 减去一个月
        if current_month == 1:
            current_month = 12
            current_year -= 1
        else:
            current_month -= 1

    # 得到所需日期
    _date = datetime( current_year, current_month, 1 )

    return _date
